import os
from stat import S_IFBLK

from .size import ByteSize


class Device(object):
    def __init__(self, path):
        # strip all symbolic links and make path absolute (for /proc/mounts)
        self.path = os.path.abspath(os.path.realpath(path))
        self.st = os.lstat(self.path)

    @property
    def is_device(self):
        return bool(self.st.st_mode & S_IFBLK)

    @property
    def major(self):
        return os.major(self.st.st_rdev)

    @property
    def minor(self):
        return os.minor(self.st.st_rdev)

    @property
    def sys_fs_path(self):
        return os.path.join('/sys/dev/block', '{0.major}:{0.minor}'
                            .format(self))

    def _lookup_sys(self, name):
        return open(
            os.path.join(self.sys_fs_path, name), 'rb'
        ).read().rstrip('\n')

    def _lookup_sys_bool(self, name):
        return int(self._lookup_sys(name)) == 1

    @property
    def removable(self):
        return self._lookup_sys_bool('removable')

    @property
    def size(self):
        return ByteSize(int(self._lookup_sys('size')) * 512)

    @property
    def model(self):
        return self._lookup_sys('device/model')

    @property
    def read_only(self):
        return self._lookup_sys_bool('ro')

    def open(self, mode='r'):
        return open(self.path, mode)

    def __repr__(self):
        return '{0.__class__.__name__}({0.path})'.format(self)

    @classmethod
    def iter_block_devices(cls, base_path='/dev'):
        for name in os.listdir(base_path):
            if name.startswith('sd') and len(name) == 3:
                yield cls(os.path.join(base_path, name))



from functools import partial
import os
from stat import S_IFREG
import sys
import time

import click

from .devices import Device
from .size import ByteSize


# this will possibly have to be adjusted in the future with huge usb-sticks
G = 1024 ** 3
MAX_SIZE = 32 * G  # 32G


def candidate_for_writing(dev):
    if not dev.is_device:
        return False, 'not a device'

    if not dev.major == 8:
        return False, 'major device number is not 8 (SCSI)'

    if dev.read_only:
        return False, 'device is read-only'

    if not dev.removable:
        return False, 'device is not a removable device'

    if dev.size == 0:
        return False, 'device has 0 size'

    if dev.size > MAX_SIZE:
        return False, 'device has a size larger than {} bytes'.format(MAX_SIZE)

    return True, None


def dev_info(dev):
    return '{0.path} ({0.size:.1fH}): {0.model}'.format(dev)


@click.command()
@click.argument('image-file', type=click.Path(readable=True, dir_okay=False,
                                              exists=True))
@click.option('--target', '-t',
              type=click.Path(dir_okay=False, writable=True),
              help='The target to write to. If none is given, a menu is shown'
                   ' to select one.')
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.option('--limit', '-l', type=ByteSize)
@click.option('--i-know-what-im-doing', is_flag=True, default=False)
@click.option('--chunk-size', '-C', type=ByteSize, default=None,
              help='Read-buffer size (default:Lauto-adjust)')
def wrimg(image_file, target, verbose, i_know_what_im_doing, limit,
          chunk_size):
    if verbose:
        info = click.echo
    else:
        info = lambda *args, **kwargs: None

    error = partial(click.echo, err=True)

    if not target:
        candidates = []
        for dev in sorted(Device.iter_block_devices(), key=lambda d: d.path):
            ok, msg = candidate_for_writing(dev)
            if not ok:
                info('skipping {}: {}'.format(dev.path, msg))
                continue
            candidates.append(dev)

        if not candidates:
            error('No device given on command-line and no suitable candidate '
                  'found')
            sys.exit(1)
        elif len(candidates) == 1:
            target = candidates[0]
        else:
            # display menu
            for i, c in enumerate(candidates):
                click.echo('[{}] {}'.format(i, dev_info(c)))

            idx = int(click.prompt(
                'Select a device',
                type=click.Choice(map(str, range(len(candidates))))
            ))

            target = candidates[idx]
    else:
        target = Device(target)

    # sanity-check target-device
    ok, msg = candidate_for_writing(target)

    if not ok:
        if not i_know_what_im_doing:
            error('{.path}: {}.\nAdd --i-know-what-im-doing to disable this '
                  'check.'
                  .format(target, msg))
            sys.exit(0)
        error('WARNING: {}: {}'.format(dev_info(target), msg))

    # confirm start, featuring the defensive assert!
    assert click.confirm('Write to {}?'.format(dev_info(target)),
                         err=True, abort=True)

    # determine number of bytes to write
    img_st = os.stat(image_file)

    # start with the device size
    total = target.size

    if img_st.st_mode & S_IFREG:
        # regular file, limit by its size
        total = min(total, img_st.st_size)

    # apply limt
    if limit is not None:
        total = min(total, limit)

    info('Copying {:.0f} bytes from {} to {.path}'.format(
         total, image_file, target))

    # start with 4K, go up to 4M with bufsize
    MIN_BUFSIZE = 4 * 1024
    MAX_BUFSIZE = 1024 * MIN_BUFSIZE
    bufsize = chunk_size or MIN_BUFSIZE

    with click.progressbar(length=total, label='writing', info_sep=' | ')\
            as bar, open(image_file, 'rb') as src, target.open('wb') as dst:
        while total:
            # measure time
            start = time.time()

            buf = src.read(min(total, bufsize))
            if not buf:
                break
            dst.write(buf)
            dst.flush()
            os.fsync(dst.fileno())

            end = time.time()

            speed = ByteSize(len(buf) / (end-start))

            # adjust chunk size if needed
            if chunk_size is None:
                if end - start > 2 and bufsize > MIN_BUFSIZE:
                    # took longer then two seconds, halve bufsize
                    bufsize //= 2
                elif end - start < 0.5 and bufsize < MAX_BUFSIZE:
                    bufsize *= 2

            total -= len(buf)
            bar.label = '{:.1f H}/s'.format(speed)
            bar.update(len(buf))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='wrimg',
    version='0.1',
    description='An alternative to dd for writing images to sd-cards',
    long_description=read('README.rst'),
    author='Marc Brinkmann',
    author_email='git@marcbrinkmann.de',
    url='http://github.com/mbr/wrimg',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    install_requires=['click'],
    entry_points={
        'console_scripts': [
            'wrimg = wrimg.cli:wrimg',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 2',
        # python3 support should be possible, but currently untested
        #'Programming Language :: Python :: 3',
    ]
)



import re

UNITS_1000 = ['kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte',
              'exabyte', 'zetabyte', 'yottabyte']
UNITS_1024 = ['kibibyte', 'mebibyte', 'gibibyte', 'tebibyte', 'pebibyte',
              'exbibyte', 'zebibyte', 'yobibyte']
U_1000 = ['k', 'm', 'g', 't', 'p', 'e', 'z', 'y']
U_1024 = ['K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']

_PARSE = [
    (UNITS_1000, 1000),
    (UNITS_1024, 1024),
    (U_1000, 1000),
    (U_1024, 1024),
]

FMT_RE = re.compile(r'(.*?)( )?(hh|HH|h|H)$')
VAL_RE = re.compile(r'(\d+(?:\.\d+))(?: )?(.*)')


class ByteSize(long):
    @staticmethod
    def __new__(cls, val):
        if isinstance(val, (str, unicode)):
            val = str(val)

            for units, base in _PARSE:
                for idx, suffix in enumerate(units):
                    if val.endswith(suffix):
                        val = val[:-len(suffix)]
                        num = long(val) * base ** (idx+1)
                        return long.__new__(cls, num)
            return long(val)
        else:
            # if no string, just pass through
            return long.__new__(cls, val)

    def __format__(self, fmt):
        base = None

        m = FMT_RE.match(fmt)
        space = ''

        if m:
            fmt, space, bf = m.groups()
            if bf == 'hh':
                base = 1000
                units = UNITS_1000
            elif bf == 'HH':
                base = 1024
                units = UNITS_1024
            elif bf == 'h':
                base = 1000
                units = U_1000
            elif bf == 'H':
                base = 1024
                units = U_1024

        val = self
        suffix = ''

        if base:
            for idx, name in enumerate(units):
                if abs(val) < base:
                    break

                val /= float(base)
                suffix = units[idx]

        return float(val).__format__(fmt) + (space or '') + suffix
